package main

import (
	"strings"

	"mvdan.cc/sh/v3/syntax"
)

var ignored_lines = []uint{}

func get_ignored_and_deps(code string) ([]string, []string) {
	// source: https://www.gnu.org/software/bash/manual/html_node/Bash-Builtins.html
	var ignored []string = []string{"alias", "bind", "builtin", "caller", "command", "declare", "echo", "enable", "let", "local", "logout", "mapfile", "printf", "read", "readarray", "source", "type", "typeset", "ulimit", "unalias"}
	var deps []string

	for i, line := range strings.Split(code, "\n") {
		splitted := strings.Split(line, " ")

		// comments w/ no arguments
		if splitted[0] == "#bshchk:ignore-next-line" {
			ignored_lines = append(ignored_lines, uint(i)+1)
		}

		// comments with arguments
		if len(splitted) < 2 {
			continue
		}
		if splitted[0] == "#bshchk:ignore-cmd" {
			ignored = append(ignored, splitted[1:]...)
		}
		if splitted[0] == "#bshchk:add-cmd" {
			deps = append(deps, splitted[1:]...)
		}
	}

	return ignored, deps
}

func is_line_ignored(line uint) bool {
	for _, ignored := range ignored_lines {
		if line == ignored {
			return true
		}
	}
	return false
}

func find(code string) ([]string, error) {
	r := strings.NewReader(code)
	f, err := syntax.NewParser().Parse(r, "")
	if err != nil {
		return make([]string, 0), err
	}

	ignored, deps := get_ignored_and_deps(code)

	// 1. find function declarations
	syntax.Walk(f, func(node syntax.Node) bool {
		switch x := node.(type) {
		case *syntax.FuncDecl:
			ignored = append(ignored, x.Name.Value)
		}

		return true
	})

	// 2. collect all commands
	syntax.Walk(f, func(node syntax.Node) bool {
		switch x := node.(type) {
		case *syntax.CallExpr:
			if is_line_ignored(node.Pos().Line() - 1) {
				return true
			}
			for i := range x.Args {
				for _, part := range x.Args[i].Parts {
					switch xx := part.(type) {
					case *syntax.Lit:
						deps = append(deps, xx.Value)
					}
					return true
				}
			}
		}

		return true
	})

	var finished_deps []string

	for _, dep := range deps {
		is_builtin := false
		for _, builtin := range ignored {
			if dep == builtin {
				is_builtin = true
			}
		}

		if !is_builtin {
			finished_deps = append(finished_deps, dep)
		}
	}

	return finished_deps, nil
}
